@@ -1,3 +1,52 @@
+""" Python wrapper for running a compiler
+
+Important classes:
+    - `SourceProgram`: represents a C or C++ program together with additional
+      info (flags, macro definitions, includ paths) that can be compiled.
+    - `CompilerExe`: a compiler executable (e.g., "gcc", "/usr/bin/clang-14", etc)
+    - `CompilationOutput`: the output of a compiler invocation, e.g., an object file
+    - `CompilationSetting`: a CompilerExe together with an optimization level and
+      additional flags used to compile `SourceProgram`s. Can be parsed from a
+      string, "gcc -O3 test.c -o test.o -MACRO1=foo -fomit-frame-pointer"  with
+      `parse_compilation_setting_from_string`
+
+Example:
+from pathlib import Path
+from diopter.compiler import (
+    CompilationSetting,
+    CompilerExe,
+    CompilerProject,
+    ExeCompilationOutput,
+    Language,
+    OptLevel,
+    SourceProgram,
+)
+input_code = \"\"\"
+             #include <stdio.h>
+             void foo(int argc){
+                 printf("%d \\n", argc);
+             }
+             int main(int argc, char* argv[]){
+                 foo(argc);
+             }
+\"\"\"
+# Create a program
+program = SourceProgram(code=input_code, language=Language.C)
+# Get the system gcc compiler
+compiler = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
+# Create a compilation setting
+cs = CompilationSetting(compiler=compiler, opt_level=OptLevel.O2)
+# Compile the program to an executable
+res = cs.compile_program( program, ExeCompilationOutput())
+# Run the program
+output = res.output.run()
+# The output should be 1 (argc is 1 if no flags a passed)
+assert output.stdout.strip() == "1"
+# Run the program with one flag
+output = res.output.run(("-flag",))
+# The output should be 2
+assert output.stdout.strip() == "2"
+"""
 from __future__ import annotations
 
 import argparse
@@ -5,15 +54,16 @@ import os
 import re
 import subprocess
 import tempfile
+from abc import ABC, abstractmethod
 from dataclasses import dataclass, replace
 from enum import Enum
 from itertools import chain
 from pathlib import Path
 from shutil import which
-from types import TracebackType
-from typing import TypeVar
+from subprocess import Popen
+from typing import IO, Any, Generic, TypeVar
 
-from diopter.utils import run_cmd
+from diopter.utils import CommandOutput, run_cmd, run_cmd_async, temporary_file
 
 
 class Language(Enum):
@@ -335,63 +385,366 @@ class CompileError(Exception):
         return CompileError(output)
 
 
-class TemporaryFile:
-    def __init__(self, *, contents: str, suffix: str):
-        self.contents = contents
-        self.suffix = suffix
-        self.path: str | None = None
+class CompilationOutput(ABC):
+    """Represents the output of a compiler invocation,e.g.,
+    an Object file, an Executable, assembly code, etc.
 
-    def __enter__(self) -> Path:
-        fd, self.path = tempfile.mkstemp(suffix=self.suffix)
-        os.close(fd)
-        if self.contents:
-            with open(self.path, "w") as f:
-                f.write(self.contents)
-        return Path(self.path).resolve(strict=True)
+    Subclasses of CompilationOutput are used to specify what kind of output is
+    desired in `CompilationResult.compile_program`.
 
-    def __exit__(
-        self,
-        exc_type: type[BaseException] | None,
-        exc_value: BaseException | None,
-        exc_traceback: TracebackType | None,
-    ) -> None:
-        if self.path and Path(self.path).exists():
-            os.remove(self.path)
+    Attributes:
+        filename (Path):
+            the output filename that will be passed as `-o filename` to the
+            compiler invocation, if not specified a temporary file will be
+            created.
+        temporary_file (tempfile.NamedTemporaryFile):
+            a temporary file where the output will be writen, if no filename
+            is specified when creating the CompilationOutput object
+
+    """
+
+    def __init__(self, filename: Path | None = None) -> None:
+        """Create a compilation output, either using the specified path or with
+        a temporary file. The temporary file is automatically removed with the
+        CompilationOutput object is garbage collected.
+
+        Args:
+            filename (Path | None):
+                where to write the compilation output if not None
+        """
+
+        if filename:
+            self.filename_ = filename
+            self.temporary_file = None
+        else:
+            self.temporary_file = tempfile.NamedTemporaryFile(
+                suffix=type(self).suffix(), delete=False
+            )
+            self.temporary_file.close()
+            self.filename_ = Path(self.temporary_file.name)
+
+    @property
+    def filename(self) -> Path:
+        return self.filename_
+
+    def __del__(self) -> None:
+        """Delete the temporary file if one was created and still exists"""
+        if self.temporary_file is None:
+            return
+        if Path(self.filename).exists():
+            os.remove(self.filename)
+
+    def to_cmd(self) -> str:
+        """Create the relevant compilation flags for this output.
+
+        Used in CompilationSetting.get_compilation_cmd.
+
+        Returns:
+            str:
+                the necessary compilation flags, e.g., "-c -o filename.o"
+        """
+        if type(self).empty_command():
+            return ""
+        return type(self).flag() + " -o " + str(self.filename)
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, CompilationOutput):
+            raise NotImplementedError
+        return (
+            type(self) == type(other)
+            and self.filename == other.filename
+            and self.temporary_file == other.temporary_file
+        )
+
+    @staticmethod
+    @abstractmethod
+    def flag() -> str:
+        """Return the relevant flag for this output.
+
+        E.g., for object outputs it should be "-c"
+
+        Subclasses of CompilationOutput should implement this method.
+
+        Returns:
+            str:
+                the compiler flag for this kind of output
+        """
+        raise NotImplementedError
+
+    @staticmethod
+    @abstractmethod
+    def suffix() -> str:
+        """Return the relevant filename suffix for this output.
+
+        E.g., for object outputs it should be ".o"
+
+        Subclasses of CompilationOutput should implement this method.
+
+        Returns:
+            str:
+                the filename output for the compilation output
+        """
+
+        raise NotImplementedError
+
+    @staticmethod
+    def empty_command() -> bool:
+        """Whether this kind of output should not include any compiler flags.
+
+        Subclasses of CompilationOutput can override this method.
+
+        Currently only relevant to NoCompilationOutput.
+
+        Returns:
+            bool:
+                if this compilation output kind should not inlcude any flags
+        """
+        return False
+
+
+class BinaryOutputMixin(ABC):
+    """Mixin for binary compilation outputs adding utility methods."""
+
+    @property
+    @abstractmethod
+    def filename(self) -> Path:
+        raise NotImplementedError
+
+    def strip_symbols(self) -> None:
+        """Runs the strip utility on the output, this the program's removes
+        symbols.
+
+        Useful, e.g., for comparing whether two outputs are equal.
+        """
+        run_cmd(f"strip {self.filename}")
+
+    def read(self) -> bytes:
+        """Read the output.
+
+        Returns:
+            bytes:
+                The binary's contents.
+        """
+
+        with open(str(self.filename), "rb") as f:
+            return f.read()
+
+
+class ExeCompilationOutput(CompilationOutput, BinaryOutputMixin):
+    """An executable compilation output.
+
+    Attributes: (see `CompilationOutput`)
+    """
+
+    def __init__(self, filename: Path | None = None) -> None:
+        """Create an ExeCompilationOutput and set the correct permissions
+        if a temporary file is created such that it is executable.
+        """
+        super().__init__(filename)
+        if self.temporary_file is not None:
+            os.chmod(self.filename, 0o777)
+        else:
+            # check for permissions if the file exists?
+            pass
+
+    def run(self, flags: tuple[str, ...] = tuple()) -> CommandOutput:
+        """Runs the exe with the provided flags.
+
+        Args:
+            flags (tuple[str]):
+                flags passed to the exe
+        Returns:
+            CommandOutput:
+                the captured stdout and stderr
+        """
+        return run_cmd(f"{self.filename} {' '.join(flags)}")
+
+    @staticmethod
+    def flag() -> str:
+        return ""
+
+    @staticmethod
+    def suffix() -> str:
+        return ".exe"
+
+
+class ObjectCompilationOutput(CompilationOutput, BinaryOutputMixin):
+    """An object file compilation output.
+
+    Attributes: (see `CompilationOutput`)
+    """
+
+    @staticmethod
+    def flag() -> str:
+        return "-c"
+
+    @staticmethod
+    def suffix() -> str:
+        return ".o"
+
+
+class ASMCompilationOutput(CompilationOutput):
+    """An assembly file compilation output.
+
+    Attributes: (see `CompilationOutput`)
+    """
+
+    @staticmethod
+    def flag() -> str:
+        return "-S"
+
+    @staticmethod
+    def suffix() -> str:
+        return ".s"
+
+    def read(self) -> str:
+        """Read the assembly code.
+
+        Returns:
+            str:
+                the assembly code read from the output file
+        """
+        with open(str(self.filename), "r") as f:
+            return f.read()
+
+
+class LLVMIRCompilationOutput(CompilationOutput):
+    """An LLVM IR file compilation output.
+
+    Attributes: (see `CompilationOutput`)
+    """
+
+    @staticmethod
+    def flag() -> str:
+        return "-S -emit-llvm"
+
+    @staticmethod
+    def suffix() -> str:
+        return ".ll"
+
+    def read(self) -> str:
+        """Read the LLVM IR.
+
+        Returns:
+            str:
+                the LLVM IR read from the output file
+        """
+        with open(str(self.filename), "r") as f:
+            return f.read()
+
+
+class NoCompilationOutput(CompilationOutput):
+    """No compilation output.
+
+    Used in `parse_compilation_setting_from_string`.
+    """
+
+    def __init__(self) -> None:
+        super().__init__(Path(""))
+
+    @staticmethod
+    def flag() -> str:
+        return ""
+
+    @staticmethod
+    def suffix() -> str:
+        return ""
+
+    @staticmethod
+    def empty_command() -> bool:
+        return True
+
+
+CompilationOutputType = TypeVar("CompilationOutputType", bound=CompilationOutput)
 
 
 @dataclass(frozen=True, kw_only=True)
-class CompilationInfo:
+class CompilationResult(Generic[CompilationOutputType]):
+    """A compilation result
+
+    Attributes:
+        source_file (Path):
+            the (temporary) source file used by the compiler
+        output (CompilationOutputType):
+            the output of the compiler, e.g., an executable (ExeCompilationOutput)
+        stdout_stderr_output (str):
+            the captured stdout and stderr
+    """
+
     source_file: Path
+    output: CompilationOutputType
     stdout_stderr_output: str
 
 
-class CompilationOutputKind(Enum):
-    Object = 0
-    Assembly = 1
-    Exec = 2
-    Unspecified = 3
-
-    def to_flag(self) -> str:
-        match self:
-            case CompilationOutputKind.Object:
-                return "-c"
-            case CompilationOutputKind.Assembly:
-                return "-S"
-            case CompilationOutputKind.Exec | CompilationOutputKind.Unspecified:
-                return ""
-
-
 @dataclass(frozen=True)
-class CompilationOutput:
-    filename: Path
-    kind: CompilationOutputKind
+class AsyncCompilationResult(Generic[CompilationOutputType]):
+    """A wrapper over a pending `CompilationResult`
+    from a compiler that is running in a subprocess.
 
-    def to_cmd(self) -> str:
-        if self.kind == CompilationOutputKind.Unspecified:
-            return ""
-        if kind_flag := self.kind.to_flag():
-            return kind_flag + " -o " + str(self.filename)
-        return "-o " + str(self.filename)
+    Returned by CompilationSetting.compile_program_async
+
+    Attributes:
+        cmd (str):
+            the compiler command that is running in the subprocess
+        proc (Popen[Any]):
+            the running subprocess
+        code_file (IO[bytes]):
+            the file containing the source code
+        output (CompilationOutputType):
+            the pending compilation output
+    """
+
+    cmd: str
+    proc: Popen[Any]
+    code_file: IO[bytes]
+    output: CompilationOutputType
+
+    def result(
+        self, timeout: int | None = None
+    ) -> CompilationResult[CompilationOutputType]:
+        """Waits for the subprocess to finish
+        and returns the compilation output.
+
+        Args:
+            timeout (int | None):
+                if not None, how many seconds to wait before
+                raising a subprocess.TimeoutExpired
+
+        Returns:
+            CompilationOutputType:
+                the compilation output
+        """
+        outs, errs = self.proc.communicate(timeout=timeout)
+        assert isinstance(outs, str)
+        assert isinstance(errs, str)
+
+        if self.proc.returncode != 0:
+            output = self.cmd
+            if outs:
+                output += "\nSTDOUT====\n" + outs
+            if errs:
+                output += "\nSTDERR====\n" + errs
+            raise CompileError(output)
+
+        return CompilationResult(
+            source_file=Path(self.code_file.name),
+            output=self.output,
+            stdout_stderr_output=outs + "\n" + errs,
+        )
+
+    def wait(self, timeout: int | None = None) -> None:
+        """Waits for the subprocess to finish.
+
+        Args:
+            timeout (int | None):
+                if not None, how many seconds to wait before
+                raising a subprocess.TimeoutExpired
+        """
+        self.proc.wait(timeout)
+
+    def __del__(self) -> None:
+        if not self.proc.returncode:
+            self.proc.kill()
 
 
 @dataclass(frozen=True, kw_only=True)
@@ -457,8 +810,7 @@ class CompilationSetting:
                 flags from the program are extracted and
                 the corresponding path are included in the output
             output (CompilationOutput):
-                the output path and potentially additional flags
-                based on the output kind
+                the ouput of the compilation
             include_language_flags (bool):
                 whether to include additional flags that specify
                 the source language and relevant linker flags
@@ -484,7 +836,7 @@ class CompilationSetting:
             )
         )
 
-        if include_language_flags and output.kind == CompilationOutputKind.Exec:
+        if include_language_flags and isinstance(output, ExeCompilationOutput):
             if linker_flag := program[0].language.get_linker_flag():
                 cmd.append(linker_flag)
         cmd.append(output.to_cmd())
@@ -492,129 +844,84 @@ class CompilationSetting:
 
     def compile_program(
         self,
-        program: ProgramType,
-        output: CompilationOutput,
+        program: SourceProgram,
+        output: CompilationOutputType,
         additional_flags: tuple[str, ...] = tuple(),
         timeout: int | None = None,
-    ) -> CompilationInfo:
+    ) -> CompilationResult[CompilationOutputType]:
+        """Compile a program with this setting.
 
-        with TemporaryFile(
+        Args:
+            program (ProgramType):
+                input program
+            output (CompilationOutputType):
+                the desired output, e.g., executable or object file
+            additional_flags (tuple[str, ...]):
+                additional flags used for the compilation
+            timeout (int | None):
+                timeout in seconds for the compilation command
+
+        Returns:
+            CompilationResult[CompilationOutputType]:
+                The result of the compilation (if successful).
+        """
+        code_file = temporary_file(
             contents=program.get_modified_code(), suffix=program.get_file_suffix()
-        ) as code_file:
-            cmd = self.get_compilation_cmd((program, code_file), output, True) + list(
-                additional_flags
+        )
+        cmd = self.get_compilation_cmd(
+            (program, Path(code_file.name)), output, True
+        ) + list(additional_flags)
+        try:
+            command_output = run_cmd(
+                cmd,
+                timeout=timeout,
+                additional_env={"TMPDIR": str(tempfile.gettempdir())},
             )
-            try:
-                command_output = run_cmd(
-                    cmd,
-                    timeout=timeout,
-                    additional_env={"TMPDIR": str(tempfile.gettempdir())},
-                )
-            except subprocess.CalledProcessError as e:
-                raise CompileError.from_called_process_exception(" ".join(cmd), e)
-            return CompilationInfo(
-                source_file=Path(code_file),
-                stdout_stderr_output=command_output.stdout
-                + "\n"
-                + command_output.stderr,
-            )
-
-    def get_llvm_ir_from_program(
-        self, program: ProgramType, timeout: int | None = None
-    ) -> str:
-        """Extracts LLVM-IR from the program.
-
-        This only works with LLVM compilers.
-
-        Args:
-           program (ProgramType): input program
-           timeout (int | None): timeout in seconds for the compilation command
-        Returns:
-            str:
-                LLVM-IR
-        """
-        return self.get_asm_from_program(program, ("--emit-llvm",), timeout=timeout)
-
-    def get_asm_from_program(
-        self,
-        program: ProgramType,
-        additional_flags: tuple[str, ...] = (),
-        timeout: int | None = None,
-    ) -> str:
-        """Extracts assembly code from the program.
-
-        Args:
-           program (ProgramType): input program
-           additional_flags (tuple[str, ...]): additional flags used for the compilation
-           timeout (int | None): timeout in seconds for the compilation command
-
-        Returns:
-            str:
-                assembly code
-        """
-        with TemporaryFile(contents="", suffix=".s") as asm_file:
-            self.compile_program(
-                program,
-                CompilationOutput(asm_file, CompilationOutputKind.Assembly),
-                additional_flags,
-                timeout,
-            )
-            with open(str(asm_file), "r") as f:
-                return f.read()
-
-    def compile_program_to_object(
-        self,
-        program: ProgramType,
-        object_file: Path,
-        additional_flags: tuple[str, ...] = tuple(),
-        timeout: int | None = None,
-    ) -> CompilationInfo:
-        """Compiles program to object file
-
-        Args:
-           program (ProgramType): input program
-           object_file (Path): path to the output
-           additional_flags (tuple[str, ...]): additional flags used for the compilation
-           timeout (int | None): timeout in seconds for the compilation command
-
-        Returns:
-            CompilationInfo:
-                information about the compilation
-        """
-
-        return self.compile_program(
-            program,
-            CompilationOutput(object_file, CompilationOutputKind.Object),
-            additional_flags,
-            timeout=timeout,
+        except subprocess.CalledProcessError as e:
+            raise CompileError.from_called_process_exception(" ".join(cmd), e)
+        return CompilationResult(
+            source_file=Path(code_file.name),
+            output=output,
+            stdout_stderr_output=command_output.stdout + "\n" + command_output.stderr,
         )
 
-    def compile_program_to_executable(
+    def compile_program_async(
         self,
-        program: ProgramType,
-        executable_path: Path,
+        program: SourceProgram,
+        output: CompilationOutputType,
         additional_flags: tuple[str, ...] = tuple(),
-        timeout: int | None = None,
-    ) -> CompilationInfo:
-        """Compiles program to object file
+    ) -> AsyncCompilationResult[CompilationOutputType]:
+        """Compile a program with this setting asynchronously.
+
+        You most likely want `compile_program` instead of this method.
+        `compile_program_async` is useful if the calling process must
+        interact with the compiler, e.g., via IPC
 
         Args:
-           program (ProgramType): input program
-           executable_file (Path): path to the output
-           additional_flags (tuple[str, ...]): additional flags used for the compilation
-           timeout (int | None): timeout in seconds for the compilation command
+            program (ProgramType):
+                input program
+            output (CompilationOutputType):
+                the desired output, e.g., executable or object file
+            additional_flags (tuple[str, ...]):
+                additional flags used for the compilation
 
         Returns:
-            CompilationInfo:
-                information about the compilation
+            AsyncCompilationResult[CompilationOutputType]:
+                The result of the compilation (if successful).
         """
-
-        # TODO: check/set file permissions of executable_path if it exists?
-        return self.compile_program(
-            program,
-            CompilationOutput(executable_path, CompilationOutputKind.Exec),
-            additional_flags,
-            timeout=timeout,
+        code_file = temporary_file(
+            contents=program.get_modified_code(), suffix=program.get_file_suffix()
+        )
+        cmd = self.get_compilation_cmd(
+            (program, Path(code_file.name)), output, True
+        ) + list(additional_flags)
+        return AsyncCompilationResult(
+            " ".join(cmd),
+            run_cmd_async(
+                cmd, additional_env={"TMPDIR": str(tempfile.gettempdir())}, text=True
+            ),
+            code_file,
+            output,
         )
 
     def preprocess_program(
@@ -627,25 +934,28 @@ class CompilationSetting:
         """Preprocesses the program
 
         Args:
-           program (ProgramType): input program
-           additional_flags (tuple[str, ...]): additional flags used for the compilation
-           make_compiler_agnostic (bool):
-               if true will try to remove certain constructs (e.g., attributes)
-               such that the resulting program can be compiled with both gcc and clang
-           timeout (int | None): timeout in seconds for the compilation command
+            program (ProgramType):
+                input program
+            additional_flags (tuple[str, ...]):
+                additional flags used for the compilation
+            make_compiler_agnostic (bool):
+                if true will try to remove certain constructs (e.g., attributes)
+                such that the resulting program can be compiled with both gcc and clang
+            timeout (int | None):
+                timeout in seconds for the compilation command
 
         Returns:
-            Source:
+            ProgramType:
                 the prepocessed program
-
         """
 
-        preprocessed_source = self.compile_program(
+        result = self.compile_program(
             program,
-            CompilationOutput(Path(""), CompilationOutputKind.Unspecified),
+            ASMCompilationOutput(None),
             ("-P", "-E") + additional_flags,
             timeout=timeout,
-        ).stdout_stderr_output
+        )
+        preprocessed_source = result.output.read()
 
         if make_compiler_agnostic:
             # remove malloc attributes with args, clang doesn't understand these
@@ -660,6 +970,13 @@ class CompilationSetting:
             preprocessed_source = re.sub(
                 r"typedef [^;]*_Float\d+x?;", r"", preprocessed_source
             )
+            # replace remaining FloatX types with the standard ones
+            preprocessed_source = re.sub(r"_Float32x", r"double", preprocessed_source)
+            preprocessed_source = re.sub(
+                r"_Float64x", r"long double", preprocessed_source
+            )
+            preprocessed_source = re.sub(r"_Float32", r"float", preprocessed_source)
+            preprocessed_source = re.sub(r"_Float64", r"double", preprocessed_source)
 
         return program.with_preprocessed_code(preprocessed_source)
 
@@ -680,24 +997,24 @@ def find_standard_include_paths(
         tuple[str,...]:
             the include paths
     """
-    with TemporaryFile(contents="", suffix=".c" if not cpp else ".cpp") as tf:
-        # run clang with verbose output on an empty temporary file
-        cmd = [str(clang.exe), str(tf), "-c", "-o/dev/null", "-v"]
-        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
-        assert result.returncode == 0
+    tf = temporary_file(contents="", suffix=".c" if not cpp else ".cpp")
+    # run clang with verbose output on an empty temporary file
+    cmd = [str(clang.exe), str(tf.name), "-c", "-o/dev/null", "-v"]
+    result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+    assert result.returncode == 0
 
-        # parse the output and extract includes
-        output = result.stdout.decode("utf-8").split("\n")
-        start = (
-            next(
-                i
-                for i, line in enumerate(output)
-                if "#include <...> search starts here:" in line
-            )
-            + 1
+    # parse the output and extract includes
+    output = result.stdout.decode("utf-8").split("\n")
+    start = (
+        next(
+            i
+            for i, line in enumerate(output)
+            if "#include <...> search starts here:" in line
         )
-        end = next(i for i, line in enumerate(output) if "End of search list." in line)
-        return tuple(output[i].strip() for i in range(start, end))
+        + 1
+    )
+    end = next(i for i, line in enumerate(output) if "End of search list." in line)
+    return tuple(output[i].strip() for i in range(start, end))
 
 
 class ClangToolMode(Enum):
@@ -777,52 +1094,52 @@ class ClangTool:
 
         """
 
-        with TemporaryFile(
+        tf = temporary_file(
             contents=program.get_modified_code(), suffix=program.get_file_suffix()
-        ) as tf:
-            match program.language:
-                case Language.C:
-                    standard_include_paths = self.standard_c_include_paths
-                case Language.CPP:
-                    standard_include_paths = self.standard_cxx_include_paths
+        )
+        match program.language:
+            case Language.C:
+                standard_include_paths = self.standard_c_include_paths
+            case Language.CPP:
+                standard_include_paths = self.standard_cxx_include_paths
 
-            cmd = list(
-                chain(
-                    (str(self.exe), str(tf)),
-                    tool_flags,
-                    (f"--extra-arg=-isystem{path}" for path in standard_include_paths),
-                    (f"--extra-arg={flag}" for flag in program.get_compilation_flags()),
-                    ("--",),
-                )
+        cmd = list(
+            chain(
+                (str(self.exe), str(tf.name)),
+                tool_flags,
+                (f"--extra-arg=-isystem{path}" for path in standard_include_paths),
+                (f"--extra-arg={flag}" for flag in program.get_compilation_flags()),
+                ("--",),
             )
-            try:
-                result = run_cmd(
-                    cmd,
-                    timeout=timeout,
-                    additional_env={"TMPDIR": str(tempfile.gettempdir())},
-                )
-            except subprocess.CalledProcessError as e:
-                raise CompileError.from_called_process_exception(" ".join(cmd), e)
+        )
+        try:
+            result = run_cmd(
+                cmd,
+                timeout=timeout,
+                additional_env={"TMPDIR": str(tempfile.gettempdir())},
+            )
+        except subprocess.CalledProcessError as e:
+            raise CompileError.from_called_process_exception(" ".join(cmd), e)
 
-            match mode:
-                case ClangToolMode.CAPTURE_OUT_ERR:
+        match mode:
+            case ClangToolMode.CAPTURE_OUT_ERR:
+                return ClangToolResult(
+                    modified_source_code=None,
+                    stdout=result.stdout,
+                    stderr=result.stderr,
+                )
+            case ClangToolMode.READ_MODIFIED_FILE:
+                with open(str(tf), "r") as f:
                     return ClangToolResult(
-                        modified_source_code=None,
+                        modified_source_code=f.read(), stdout=None, stderr=None
+                    )
+            case ClangToolMode.CAPTURE_OUT_ERR_AND_READ_MODIFIED_FILED:
+                with open(str(tf), "r") as f:
+                    return ClangToolResult(
+                        modified_source_code=f.read(),
                         stdout=result.stdout,
                         stderr=result.stderr,
                     )
-                case ClangToolMode.READ_MODIFIED_FILE:
-                    with open(str(tf), "r") as f:
-                        return ClangToolResult(
-                            modified_source_code=f.read(), stdout=None, stderr=None
-                        )
-                case ClangToolMode.CAPTURE_OUT_ERR_AND_READ_MODIFIED_FILED:
-                    with open(str(tf), "r") as f:
-                        return ClangToolResult(
-                            modified_source_code=f.read(),
-                            stdout=result.stdout,
-                            stderr=result.stderr,
-                        )
 
 
 @dataclass(frozen=True, kw_only=True)
@@ -867,33 +1184,31 @@ class CComp:
         # ccomp doesn't like these
         code = re.sub(r"__asm__ [^\)]*\)", r"", program.get_modified_code())
 
-        with TemporaryFile(contents=code, suffix=".c") as tf:
-            cmd = (
-                [
-                    str(self.exe),
-                    str(tf),
-                    "-interp",
-                    "-fall",
-                ]
-                + [
-                    f"-I{ipath}"
-                    for ipath in chain(
-                        program.include_paths, program.system_include_paths
-                    )
-                ]
-                + [f"-D{macro}" for macro in program.defined_macros]
+        tf = temporary_file(contents=code, suffix=".c")
+        cmd = (
+            [
+                str(self.exe),
+                str(tf.name),
+                "-interp",
+                "-fall",
+            ]
+            + [
+                f"-I{ipath}"
+                for ipath in chain(program.include_paths, program.system_include_paths)
+            ]
+            + [f"-D{macro}" for macro in program.defined_macros]
+        )
+        try:
+            run_cmd(
+                cmd,
+                additional_env={"TMPDIR": str(tempfile.gettempdir())},
+                timeout=timeout,
             )
-            try:
-                run_cmd(
-                    cmd,
-                    additional_env={"TMPDIR": str(tempfile.gettempdir())},
-                    timeout=timeout,
-                )
-            except subprocess.CalledProcessError as e:
-                if debug:
-                    print(CompileError.from_called_process_exception(" ".join(cmd), e))
-                return False
-            return True
+        except subprocess.CalledProcessError as e:
+            if debug:
+                print(CompileError.from_called_process_exception(" ".join(cmd), e))
+            return False
+        return True
 
 
 def __compilation_setting_parser() -> argparse.ArgumentParser:
@@ -979,16 +1294,16 @@ def parse_compilation_setting_from_string(
         macro_definitions=macro_definitions,
         flags=tuple(flags),
     )
-
+    output: CompilationOutput
     if parsed_args.o:
         if parsed_args.c:
-            kind = CompilationOutputKind.Object
+            output = ObjectCompilationOutput(Path(parsed_args.o))
         elif parsed_args.S:
-            kind = CompilationOutputKind.Assembly
+            # TODO: check for emit-llvm?
+            output = ASMCompilationOutput(Path(parsed_args.o))
         else:
-            kind = CompilationOutputKind.Exec
-        output = CompilationOutput(parsed_args.o, kind)
+            output = ExeCompilationOutput(Path(parsed_args.o))
     else:
-        output = CompilationOutput(Path(""), CompilationOutputKind.Unspecified)
+        output = NoCompilationOutput()
 
     return csetting, sources, output
